from .basic import (
    BasicFilter,
    NegatableFilter,
    NotEmptyFilter,
    NullishFilter,
    NumericFilter,
    StringContainsFilter,
    StringEndsFilter,
    StringFilter,
    StringStartsFilter,
    ValuesFilter,
)
from .combining import FilterSet
from .types import Combinator, GenericFilter



from __future__ import annotations

import abc
from collections.abc import MutableMapping
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    ClassVar,
    Generic,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from weakref import WeakValueDictionary

from zucker.exceptions import UnsavedRecordError

from ..client import AsyncClient, SyncClient
from ..filtering import GenericFilter
from ..utils import JsonMapping, JsonType, MutableJsonMapping
from .fields.base import Field
from .fields.scalars import IdField
from .view import AsyncView, SyncView, View

if TYPE_CHECKING:
    from ..client import BaseClient  # noqa: F401

ClientType = TypeVar("ClientType", bound="BaseClient", covariant=True)
NativeType = TypeVar("NativeType")
BoundSelf = TypeVar("BoundSelf", bound="BoundModule[BaseClient]")
SyncSelf = TypeVar("SyncSelf", bound="SyncModule")
AsyncSelf = TypeVar("AsyncSelf", bound="AsyncModule")

SyncOrAsync = Union[None, Awaitable[None]]


class BaseModule:
    """Base class for all modules.

    This class can be used to create unbounded base module types that can be reused as
    supertypes for actual models. To create actually usable module types, use the
    ``Module`` property of a client as the superclass.

    Module classes define the data model and are referred to as *records* when
    instantiated.
    """

    id = IdField()

    def __init__(self, **data: JsonType):
        self._original_data: JsonMapping = {}
        self._updated_data: MutableJsonMapping = {}
        self._set_data(data)

    #############################
    # Record instance - general #
    #############################

    def _repr_data(self) -> Sequence[str]:
        try:
            name = self["name"]
            assert isinstance(name, str)
            name_repr = f"- {name}"
        except KeyError:
            name_repr = ""
        id_repr = self.get_data("id", "")
        assert isinstance(id_repr, str)
        return (id_repr, name_repr)

    def __repr__(self) -> str:
        repr_data = (self.__class__.__name__, "record", *self._repr_data())
        return f"<{' '.join(repr_data)}>"

    def __eq__(self, other: Any) -> bool:
        """Test if this record reference is equal to another.

        Records are treated equal if their module and id match. When comparing two
        records that could potentially be in different modules, always compare the
        Record objects directly and not just their id properties.
        """
        if isinstance(other, type(self)):
            try:
                if self._id != other._id:
                    return False
            except KeyError:
                pass
            else:
                return True
        return False

    #####################################
    # Record instance - data management #
    #####################################

    def __getitem__(self, item: str) -> JsonType:
        if not isinstance(item, str):
            raise TypeError("module keys must be strings")

        try:
            value = self._updated_data[item]
            return value
        except (KeyError, TypeError):
            try:
                value = self._original_data[item]
                return value
            except (KeyError, TypeError):
                pass

        # This 'raise' statement is moved below (even though logically, it would fit
        # inside the inner 'except' in order to keep stack traces clean.
        raise KeyError(item)

    def _set_data(self, data: JsonMapping) -> None:
        self._original_data = {
            key: value for key, value in data.items() if not key.startswith("_")
        }
        self._updated_data = {}

    def get_data(
        self,
        key: str,
        default: Optional[NativeType] = None,
    ) -> Union[None, NativeType, JsonType]:
        """Get the current value of a specific key.

        This value may either come from the existing data populated when initializing
        the record or from any fields that have been updated / changed in the meantime.

        :param key: The key to look up.
        :param default: Default value to return when no value is found. If this is given
            and a value is found that doesn't match the default's type the default value
            is also returned.
        """
        try:
            value = self[key]
            if default is not None and not isinstance(value, type(default)):
                return default
            return value
        except KeyError:
            return default

    @property
    def _id(self) -> str:
        value = self["id"]
        if not isinstance(value, str):
            raise TypeError(f"got record ID of invalid type {type(value)!r}")
        return value

    ######################
    # Module information #
    ######################

    @classmethod
    def field_names(cls) -> Iterator[str]:
        """Iterate over all requested field names."""
        for key in dir(cls):
            if key.startswith("_"):
                continue
            if isinstance(getattr(cls, key, None), Field):
                yield key


class UnboundModule(BaseModule):
    """Unbound modules can be used as a base class for defining models that aren't
    connected to a client yet.
    """


class BoundModule(Generic[ClientType], BaseModule, abc.ABC):
    """Bound modules are module classes are already scoped to a client and therefore
    also to the sync or async paradigm.

    Bound modules contain all the logic that enables server-side communication. That
    means that bound records (records are the instances of module classes) can be
    saved, refreshed and deleted.
    """

    _CLIENT_TYPE: ClassVar[Type[BaseClient]]
    _client: ClassVar[BaseClient]
    _api_name: ClassVar[str]

    # Cache object that allows direct access to records by their ID. References here are
    # held weakly - that means that once the corresponding view is garbage collected,
    # entries in this cache may no longer be accessible.
    _record_cache: ClassVar[MutableMapping[str, BoundModule[BaseClient]]]

    def __init_subclass__(
        cls,
        *,
        client: Optional[ClientType] = None,
        api_name: Optional[str] = None,
    ):
        if not abc.ABC in cls.__bases__:
            if not isinstance(client, cls._CLIENT_TYPE):
                raise TypeError(
                    f"expecting client of type {cls._CLIENT_TYPE!r}, got "
                    f"{type(client)!r}"
                )
            cls._client = client
            cls._api_name = api_name or cls.__name__
            cls._record_cache = WeakValueDictionary()

    @classmethod
    def get_client(cls) -> ClientType:
        """Client instance bounded to this module.

        This client will be used for all server-side communication. Further, any caches
        are scoped to this client.
        """
        assert isinstance(cls._client, cls._CLIENT_TYPE)
        return cast(ClientType, cls._client)

    @property
    def client(self) -> ClientType:
        """Shorthand for :meth:`get_client`."""
        return self.get_client()

    #############################
    # Record instance - general #
    #############################

    def _repr_data(self) -> Sequence[str]:
        return (f"(via {self._api_name})", *super()._repr_data())

    def __eq__(self, other: Any) -> bool:
        """Test if this record reference is equal to another.

        Records are treated equal if their module and id match. When comparing two
        records that could potentially be in different modules, always compare the
        Record objects directly and not just their id properties.
        """
        if isinstance(other, type(self)):
            if self._api_name != other._api_name:  # pragma: no cover
                # This case shouldn't ever happen (and if it does it would be undefined
                # behaviour). Since we already checked that the other model has a
                # different type the only way this can happen is when a user
                # intentionally sets the _api_name on a record instance.
                return False
            if self.client is not other.client:
                return False
        return super().__eq__(other)

    def _set_data(self, data: JsonMapping) -> None:
        if data.get("_module", self._api_name) != self._api_name:
            raise ValueError(
                f"trying to set {self.__class__.__name__} record data the wrong API "
                f"type - got {data['_module']}, expecting {self._api_name}"
            )
        super()._set_data(data)

    #######################################
    # Record instance - server operations #
    #######################################

    @property
    def _mutation_endpoint(self) -> str:
        if self.get_data("id") is None:
            raise UnsavedRecordError("cannot mutate a non-saved record")
        return f"{self._api_name}/{self._id}"

    def _prepare_save(self) -> Tuple[str, str, JsonMapping]:
        """Prepare everything for a save step that doesn't require the client yet.

        This method returns a tuple containing an HTTP method, an endpoint and a
        JSON mapping that should be passed as payload. Once the save method has called
        the corresponding client, the resulting data from the request should be passed
        to :meth:``_finalize_save``.
        """
        # Get the record id (or None). This is used to determine whether we are updating
        # an existing record or creating a new one.
        record_id = self.get_data("id")

        if record_id is None:
            data_keys = set(self._updated_data.keys()) | set(self._original_data.keys())
        else:
            # If the record is already present on the server, we only need to send the
            # updated data points.
            data_keys = set(self._updated_data.keys())

        for key in data_keys:
            if key.startswith("_"):
                raise ValueError(f"cannot save underscore-prefixed key {key!r}")
        data = {key: self[key] for key in data_keys}

        if record_id is None:
            return "post", self._api_name, data
        else:
            if "id" in data:
                # TODO This should no longer be required (and could be converted into an
                #  assert) when the IdField is actually used.
                raise ValueError("cannot change a record's ID")
            return "put", f"{self._api_name}/{record_id}", data

    def _finalize_save(self, record_data: JsonMapping) -> None:
        self._set_data(record_data)

    @abc.abstractmethod
    def save(self) -> SyncOrAsync:
        """Save all updated fields back to the server."""

    def _finalize_delete(self) -> None:
        # Merge any updated data into the original data set (because we no longer have
        # a server-side record to match).
        data_keys = set(self._updated_data.keys()) | set(self._original_data.keys())
        self._original_data = {key: self.get_data(key) for key in data_keys}
        if "id" in self._original_data:
            del self._original_data["id"]
        self._updated_data = {}

    def _finalize_refresh(self, record_data: JsonMapping) -> None:
        self._set_data(record_data)

    @abc.abstractmethod
    def delete(self) -> SyncOrAsync:
        """Delete this record.

        Calling this method will delete the record on the server and remove the ID
        information from this module record instance. Subsequent calls to :meth:`save`
        will create a new record.
        """

    @abc.abstractmethod
    def refresh(self, *, _record_data: Optional[JsonMapping] = None) -> SyncOrAsync:
        """Refresh cached data from the server.

        This will clear any updated data that has been manually saved and not set yet.
        """

    ##########################
    # Record / view building #
    ##########################

    @classmethod
    @abc.abstractmethod
    def find(
        cls: Type[BoundSelf],
        *filters: Union[JsonMapping, GenericFilter]
        # This huge generic basically amounts to: "Any view that returns either Selfs
        # and optional Selfs or the same thing, but awaitable":
    ) -> View[
        BoundSelf,
        Union[BoundSelf, Awaitable[BoundSelf]],
        Union[Optional[BoundSelf], Awaitable[Optional[BoundSelf]]],
    ]:
        """Create a view on the module.

        Any parameters passed here will be used as filters.
        """

    @classmethod
    def _lookup_record_cache(cls: Type[BoundSelf], key: str) -> Optional[BoundSelf]:
        item = cls._record_cache.get(key, None)
        if item is None:
            return None
        assert isinstance(item, cls)
        return item

    def _cache_self(self) -> None:
        self._record_cache[self._id] = self

    @classmethod
    @abc.abstractmethod
    def get_by_id(
        cls: Type[BoundSelf], key: str
    ) -> Union[Optional[BoundSelf], Awaitable[Optional[BoundSelf]]]:
        """Retrieve a record object by the ID."""


class SyncModule(BoundModule[SyncClient], abc.ABC):
    _CLIENT_TYPE = SyncClient

    @classmethod
    def find(
        cls: Type[SyncSelf], *filters: Union[JsonMapping, GenericFilter]
    ) -> SyncView[SyncSelf]:
        view = SyncView(cls, cls._api_name)
        if len(filters) > 0:
            view = view.filtered(*filters)
        return view

    @classmethod
    def get(
        cls: Type[SyncSelf], *filters: Union[JsonMapping, GenericFilter]
    ) -> Optional[SyncSelf]:
        """Return the first and only element in a view, if it is present."""
        try:
            return cls.find(*filters)[0]
        except IndexError:
            return None

    def save(self) -> None:
        method, endpoint, data = self._prepare_save()
        record_data = self.client.request(method, endpoint, json=data)
        self._finalize_save(record_data)

    def delete(self) -> None:
        self.client.request("delete", self._mutation_endpoint)
        self._finalize_delete()

    def refresh(self, *, _record_data: Optional[JsonMapping] = None) -> None:
        record_data = self.client.request("get", self._mutation_endpoint)
        self._finalize_refresh(record_data)

    @classmethod
    def get_by_id(cls: Type[SyncSelf], key: str) -> Optional[SyncSelf]:
        return cls.find().get_by_id(key)


class AsyncModule(BoundModule[AsyncClient], abc.ABC):
    _CLIENT_TYPE = AsyncClient

    @classmethod
    def find(
        cls: Type[AsyncSelf], *filters: Union[JsonMapping, GenericFilter]
    ) -> AsyncView[AsyncSelf]:
        view = AsyncView(cls, cls._api_name)
        if len(filters) > 0:
            view = view.filtered(*filters)
        return view

    @classmethod
    def get(
        cls: Type[AsyncSelf], *filters: Union[JsonMapping, GenericFilter]
    ) -> Optional[AsyncSelf]:
        """Return the first and only element in a view, if it is present."""
        try:
            return await cls.find(*filters)[0]
        except IndexError:
            return None

    async def save(self) -> None:
        method, endpoint, data = self._prepare_save()
        record_data = await self.client.request(method, endpoint, json=data)
        self._finalize_save(record_data)

    async def delete(self) -> None:
        await self.client.request("delete", self._mutation_endpoint)
        self._finalize_delete()

    async def refresh(self, *, _record_data: Optional[JsonMapping] = None) -> None:
        record_data = await self.client.request("get", self._mutation_endpoint)
        self._finalize_refresh(record_data)

    @classmethod
    async def get_by_id(cls: Type[AsyncSelf], key: str) -> Optional[AsyncSelf]:
        return await cls.find().get_by_id(key)



from typing import Mapping, Optional

from zucker.utils import JsonMapping

from .base import SyncClient


class RequestsClient(SyncClient):
    """Synchronous client implementation using `requests`_.

    .. _requests: https://docs.python-requests.org/en/latest/
    """

    def __init__(
        self,
        base_url: str,
        username: str,
        password: str,
        *,
        client_platform: str = "zucker",
        verify_ssl: bool = True,
    ):
        import requests

        super().__init__(
            base_url,
            username,
            password,
            client_platform=client_platform,
            verify_ssl=verify_ssl,
        )
        self._session = requests.Session()

    def raw_request(
        self,
        method: str,
        endpoint: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        data: Optional[JsonMapping] = None,
        json: Optional[JsonMapping] = None,
    ) -> tuple[int, JsonMapping]:
        response = self._session.request(
            method,
            f"{self.base_url}/rest/v11_5/{endpoint}",
            headers={
                "OAuth-Token": self._authentication[1],
                "Cache-Control": "no-cache",
            },
            verify=self._verify_ssl,
            params=params or {},
            data=data or {},
            json=json or {},
        )
        return response.status_code, response.json()



from typing import Any, Mapping, MutableMapping, Sequence, TypeVar, Union, cast

__all__ = [
    "JsonPrimitive",
    "JsonType",
    "JsonMapping",
    "MutableJsonMapping",
    "ApiType",
    "check_json_primitive",
    "check_json_mapping",
    "check_json",
]


# MyPy currently doesn't support recursive types, so we can't actually build a correct
# 'JSON' type yet. See here: https://github.com/python/mypy/issues/731
#
# JsonType = Union[None, bool, str, int, float, Mapping[str, "JsonType"], Sequence["JsonType"]]
JsonPrimitive = Union[None, bool, str, int, float]
JsonType = Union[JsonPrimitive, Mapping, Sequence]
NonNoneJsonType = Union[bool, str, int, float, Mapping, Sequence]

JsonMapping = Mapping[str, JsonType]
MutableJsonMapping = MutableMapping[str, JsonType]

ApiType = TypeVar("ApiType", bound=JsonType)


def check_json_primitive(data: Any) -> JsonPrimitive:
    """Check if the provided object is a valid JSON primitive and return it."""
    if not isinstance(data, (type(None), bool, str, int, float)):
        raise TypeError(f"expected JSON primitive, got f{data}")
    return data


def check_json_mapping(data: Any) -> JsonMapping:
    """Recursively check if the provided object is a valid JSON mapping and
    return it.
    """
    if not isinstance(data, Mapping):
        raise TypeError(f"expected JSON dictionary, got f{data}")
    for key, value in data.items():
        if not isinstance(key, str):
            raise TypeError(f"expected string key, got {type(key)!r}")
        check_json(value)
    return cast(JsonMapping, data)


def check_json(data: Any) -> JsonType:
    """Recursively check if a provided data object is valid JSON (in the respective
    native Python types) and return it.

    This method is used as a type guard for input that comes from the backend.
    """
    try:
        return check_json_primitive(data)
    except TypeError:
        pass

    if isinstance(data, Sequence):
        for item in data:
            check_json(item)
        return cast(JsonType, data)
    elif isinstance(data, Mapping):
        return check_json_mapping(data)

    raise TypeError(f"expected valid JSON type, got {type(data)!r}")



from __future__ import annotations

import abc
import sys
from collections.abc import AsyncIterator, Mapping, Sequence
from contextlib import contextmanager
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)
from uuid import UUID

from ..exceptions import InvalidSugarResponseError
from ..filtering import Combinator, FilterSet, GenericFilter
from ..utils import JsonMapping, JsonType, check_json_mapping

if TYPE_CHECKING:
    from ..client import BaseClient  # noqa: F401
    from .module import AsyncModule, BoundModule, SyncModule  # noqa: F401

ModuleType = TypeVar("ModuleType", bound="BoundModule[BaseClient]")
SyncModuleType = TypeVar("SyncModuleType", bound="SyncModule")
AsyncModuleType = TypeVar("AsyncModuleType", bound="AsyncModule")
GetReturn = TypeVar("GetReturn", covariant=True)
OptionalGetReturn = TypeVar("OptionalGetReturn", covariant=True)
Self = TypeVar("Self", bound="View[Any, Any, Any]")


class View(Generic[ModuleType, GetReturn, OptionalGetReturn], abc.ABC):
    """Generic view class.

    The first generic argument should be passed when initializing a view and point to
    the module's type. The other two are only used internally and set by
    :class:`SyncView` and :class:`AsyncView`. They determine the type of results that
    are returned from the abstract methods.
    """

    def __init__(self, module: Type[ModuleType], base_endpoint: str = ""):
        """Build a new view.

        :param module: A class that inherits from :class:`Module` which is used to
            construct record objects. This class also determines which fields will be
            retrieved (anything that isn't defined in the module class will be ignored
            even if the server provides it).
        :param base_endpoint: Base endpoint to start queries from (without the server
            URL, which the client provides). For a simple module query, this is the
            module's name. When querying for related records, it will be something like
            ``<module>/<id>/link/<link name>``.
        """
        self._module = module
        self._base_endpoint = base_endpoint
        self._filter: Optional[GenericFilter] = None

        # The range is the current subset of indexes in the dataset that this view
        # points to. When a sub-view is created by using the slice square brackets
        # notation, the slice is added to a list of pending slices, which are all
        # applied to the range when required. That way, we can defer evaluation as long
        # as possible and are also paradigm-agnostic.
        self._range: Optional[range] = None
        self._pending_slices = list[slice]()

        # This parameter holds the total size of the view - not just that part that is
        # targeted by the range. This is fetched by calling .prefetch_size()
        self._size: Optional[int] = None

        # The record cache for this view is different from the one in the module itself
        # in that this one
        # a) holds strong references so that it is ensured that the cache is actually
        #    useful for the entire lifetime of the view.
        # b) maps from offsets instead of IDs to the corresponding record objects. That
        #    means we can iterate over the view multiple times without needing to
        #    re-fetch data.
        # This cache is meant to be supplementary to the other one - this one is for
        # iterating, the other is for accessing objects by ID.
        self._record_cache: Dict[int, Optional[ModuleType]] = {}

    #################
    # Magic methods #
    #################

    def __repr__(self) -> str:
        module_name = self._module.__name__
        prefix = "filtered view" if self._filter else "view"

        if self._range is None:
            range_repr = ""
        else:
            range_repr_parts = ["", "", ""]
            if self._range.step < 0:
                if self._range.start != sys.maxsize - 1:
                    range_repr_parts[0] = str(self._range.start)
                if self._range.stop != -1:
                    range_repr_parts[1] = str(self._range.stop)
            else:
                if self._range.start != 0:
                    range_repr_parts[0] = str(self._range.start)
                if self._range.stop != sys.maxsize:
                    range_repr_parts[1] = str(self._range.stop)
            if self._range.step != 1:
                range_repr_parts[2] = str(self._range.step)
            # Make sure we don't get an overhang like "1::"
            if range_repr_parts[2] == "":
                range_repr_parts.pop()
            range_repr = (
                ""
                if all(part == "" for part in range_repr_parts)
                else f"[{':'.join(range_repr_parts)}]"
            )

        return f"<{prefix} on {module_name}{range_repr}>"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, View):
            return False
        if other._module is not self._module:
            return False
        return all(
            getattr(self, attr) == getattr(other, attr)
            for attr in ("_base_endpoint", "_filter")
        )

    @overload
    def __getitem__(self, item: Union[UUID, str]) -> GetReturn:
        ...

    @overload
    def __getitem__(self, item: int) -> GetReturn:
        ...

    @overload
    def __getitem__(self: Self, item: slice) -> Self:
        ...

    def __getitem__(
        self: Self, item: Union[UUID, str, int, slice]
    ) -> Union[GetReturn, Self]:
        # First case - retrieving a single record by it's ID, which may either be given
        # as a string or as a UUID object.
        if isinstance(item, (UUID, str)):
            item = str(item)
            record = self.get_by_id(item)

            # Ignore this type for now until MyPy supports generic self types. See the
            # comment in __get__ in the base Field class for details.
            return record  # type: ignore

        # Second case - retrieving a single record by it's offset in the view.
        elif isinstance(item, int):
            record = self.get_by_index(item)

            # Same typing stuff as above.
            return record  # type: ignore

        # Third case - taking a slice of the view. This will return a new subview, which
        # is evaluated lazily and won't fetch any data until asked.
        elif isinstance(item, slice):
            if any(
                not isinstance(value, (type(None), int))
                for value in (item.start, item.stop, item.step)
            ):
                raise TypeError("module views only support slicing by integers")

            with self._clone() as view:
                view._pending_slices.append(item)

            return view

        else:
            raise TypeError(
                f"items in a module view must be referenced either by ID, offset or "
                f"slice; got {type(item)!r}"
            )

    ##########################
    # Extending and chaining #
    ##########################

    def filtered(self: Self, *filters: Union[JsonMapping, GenericFilter]) -> Self:
        """Return a clone of this view which contains additional filters.

        :param filters: Additional filters for the new view. These should be provided as
            filter-type objects (for example those generated by fields). Alternatively,
            naive dictionaries (those `available`_ from Sugar) are also supported. When
            multiple filters are provided, they will be AND-ed together.

        .. _available: https://support.sugarcrm.com/Documentation/Sugar_Developer/Sugar_Developer_Guide_9.0/Integration/Web_Services/REST_API/Endpoints/module_GET/#Filter_Expressions
        """
        with self._clone() as view:
            if len(filters) == 1 and isinstance(filters[0], GenericFilter):
                view._filter = filters[0]
            else:
                view._filter = FilterSet(Combinator.AND, self._filter, *filters)
        return view

    def reversed(self: Self) -> Self:
        """Return a clone of this view, but with reversed iteration order.

        This is identical to ``view[::-1]``.
        """
        return self[::-1]

    @contextmanager
    def _clone(self: Self) -> Iterator[Self]:
        """Make a new view that builds on this one, while trying to preserve the cache
        as much as possible.

        This is intended to be used as a context manager:

        .. code-block:: python

            with self._clone() as view:
                view._filter = ...
            return view

        After closing the context manager, the view will be finalized. This step tries
        to copy as much of the record cache as possible.
        """
        view = self.__class__(self._module, self._base_endpoint)
        view._filter = self._filter
        view._range = self._range
        view._pending_slices = self._pending_slices[:]

        try:
            yield view
        finally:
            # When the filter changes, we need a clean cache anyway because server-side
            # offsets may be different now.
            if view._filter != self._filter:
                return

            # In the other case, the cache can be shared between the new and old views.
            # Remember, the cache stores the server-side offsets so there shouldn't be
            # a problem with two views writing to the same cache object.
            view._record_cache = self._record_cache
            view._size = self._size

    ###################
    # Record fetching #
    ###################

    def _prepare_get_by_id(self: Self, key: str) -> Self:
        if "/" in key or " " in key:
            raise ValueError("record keys cannot contain slashes or spaces")

        return self.filtered(self._module.id == key)

    @abc.abstractmethod
    def get_by_id(self, key: str) -> GetReturn:
        """Retrieve a record object by it's ID."""

    def _prepare_get_by_offset(
        self, offset: int
    ) -> Union[ModuleType, Tuple[str, str, Mapping[str, str]]]:
        if (cache_entry := self._record_cache.get(offset, None)) is not None:
            return cache_entry

        return (
            "get",
            # https://support.sugarcrm.com/Documentation/Sugar_Developer/Sugar_Developer_Guide_11.1/Integration/Web_Services/REST_API/Endpoints/module_GET/
            f"{self._base_endpoint}",
            dict(
                max_num="1",
                offset=str(offset),
                **self._query_params,
            ),
        )

    def _finalize_get_by_offset(
        self, offset: int, data: JsonMapping
    ) -> Optional[ModuleType]:
        if "records" not in data or not isinstance(data["records"], Sequence):
            raise InvalidSugarResponseError(
                "records filter request did not return any data"
            )
        if len(data["records"]) > 1:
            raise InvalidSugarResponseError(
                f"requested a single record but got more that one response"
            )

        if len(data["records"]) == 0:
            record = None
        else:
            record_data = data["records"][0]
            try:
                record = self._module(**check_json_mapping(record_data))
            except TypeError:
                raise InvalidSugarResponseError("got invalid record data")
        self._record_cache[offset] = record
        return record

    @abc.abstractmethod
    def _get_by_offset(self, offset: int) -> OptionalGetReturn:
        """Retrieve a record object by the server-side offset, which depends on the
        current filtering.
        """

    @abc.abstractmethod
    def get_by_index(self, index: int) -> GetReturn:
        """Retrieve a record object by it's index in this view."""

    def _index_to_offset(self, index: int) -> Optional[int]:
        """Convert an offset into a server-side index.

        The difference between the two enumeration numerals is the following: indexes
        describe the actual 0-based numbering used when using *this* view as a sequence.
        It is used both when accessing it using the ``view[3]` syntax (to get the third
        record) or when iterating, for example using a ``for`` loop or something like
        ``list(view)``. These offsets differ from view to view and their ordering is
        dependent on a number of things, like:

        - The current query (filters)
        - Range the view is targeting

        Offsets on the other hand refer to the ones that the server will accept. Though
        this isn't strictly global either, it doesn't depend on the view's range any
        more. The only thing that can change offsets is when the filter gets changed.

        To sum up:

            >>> SomeModule.find()[3] # Index 3, Offset 3
            ... SomeModule.find()[3:][3] # Index 3, Offset 6
            ... SomeModule.find(some_filter)[3] # Index 3, Index depending on the filter
        """
        if index < 0:
            return None
        try:
            assert isinstance(self._range, range), "view range has not been calculated"
            return self._range[index]
        except IndexError:
            return None

    def _offset_to_index(self, offset: int) -> Optional[int]:
        try:
            assert isinstance(self._range, range), "view range has not been calculated"
            return self._range.index(offset)
        except ValueError:
            return None

    @cached_property
    def _query_params(self) -> Mapping[str, str]:
        return {
            "fields": ",".join(self._module.field_names()),
            **self._filter_query_params,
        }

    @property
    def _filter_query_params(self) -> Mapping[str, str]:
        """Render out the current filter into query parameters.

        This method will convert a filter that renders out to something like this:

        >>> {
        ...     "a": 1,
        ...     "b": [ 2, 3 ],
        ...     "c": { "d": 4 },
        ... }

        Into a dictionary like this:

        >>> {
        ...     "a": 1,
        ...     "b[0]": 2,
        ...     "b[1]": 3,
        ...     "c[d]": 4,
        ... }

        The result can be passed to an HTTP library to use as query parameters.
        """
        if self._filter is None:
            return {}

        params = dict[str, str]()

        def parse_filter(prefix: str, filter_definition: JsonType) -> None:
            """Recursively parse a filter definition and add the appropriate parameters
            to the result object."""
            if isinstance(filter_definition, (str, int, float, bool)):
                params["filter{}".format(prefix)] = str(filter_definition)
                return

            iterator: Iterable[tuple[str, JsonType]]
            if isinstance(filter_definition, Mapping):
                iterator = filter_definition.items()
            elif isinstance(filter_definition, Sequence):
                iterator = enumerate(filter_definition)  # type: ignore
            else:
                raise TypeError(
                    f"invalid filter definition type: {type(filter_definition)}"
                )

            for key, value in iterator:
                parse_filter(f"{prefix}[{key}]", value)

        parse_filter("", [self._filter.build_filter()])

        return params


class SyncView(
    Generic[SyncModuleType],
    View[SyncModuleType, SyncModuleType, Optional[SyncModuleType]],
):
    #################
    # Magic methods #
    #################

    def __len__(self) -> int:
        self._calculate_range()
        assert isinstance(self._range, range)
        return len(self._range)

    def __iter__(self) -> Iterator[SyncModuleType]:
        self._calculate_range()
        assert isinstance(self._range, range)
        iterator = iter(self._range)
        current_offset, next_offset = next(iterator, None), next(iterator, None)

        while current_offset is not None:
            # TODO Things this approach is currently missing:
            #  - Batching. At the moment, this will send a request to the server for
            #    each requested record. The batching implementation should also respect
            #    reversed ranges (when using reversed() or something like [::-1] on the
            #    view).
            #  - Dynamically determining view size, which would nullify the need for
            #    calling '/count' beforehand (which happens in __len__() in
            #    _validate_range()). This is probably only feasible for forward
            #    iteration though.
            record = self._get_by_offset(current_offset)
            if record is not None:
                yield record

            current_offset, next_offset = next_offset, next(iterator, None)

    def __reversed__(self) -> Iterator[SyncModuleType]:
        return iter(self[::-1])

    ###################
    # Record fetching #
    ###################

    def _fetch_size(self) -> None:
        """Fetch the total number of results that can be queried from this view's
        filters."""
        data = self._module.get_client().request(
            "get",
            # https://support.sugarcrm.com/Documentation/Sugar_Developer/Sugar_Developer_Guide_11.1/Integration/Web_Services/REST_API/Endpoints/modulecount_GET/
            f"{self._base_endpoint}/count",
            params=self._query_params,
        )
        record_count = data.get("record_count", None)
        if not isinstance(record_count, int):
            raise InvalidSugarResponseError(
                f"got invalid record count: {record_count!r}"
            )
        self._size = record_count

    def _calculate_range(self) -> None:
        """Apply all pending slicing operations on the cached range object."""
        if self._range is None:
            if self._size is None:
                self._fetch_size()
            assert isinstance(self._size, int)
            self._range = range(0, self._size, 1)
        while len(self._pending_slices) > 0:
            self._range = self._range[self._pending_slices.pop(0)]

    def get_by_id(self, key: str) -> SyncModuleType:
        try:
            return self._prepare_get_by_id(key)[0]
        except IndexError:
            pass
        raise KeyError(key)

    def _get_by_offset(self, offset: int) -> Optional[SyncModuleType]:
        preparation = self._prepare_get_by_offset(offset)
        if isinstance(preparation, tuple):
            method, endpoint, params = preparation
            data = self._module.get_client().request(method, endpoint, params=params)
            return self._finalize_get_by_offset(offset, data)
        else:
            return preparation

    def get_by_index(self, index: int) -> SyncModuleType:
        self._calculate_range()

        offset = self._index_to_offset(index)
        if offset is None:
            raise IndexError(index)
        record = self._get_by_offset(offset)
        if record is None:
            raise IndexError(index)
        return record


class AsyncViewIterator(Generic[AsyncModuleType]):
    def __init__(self, view: AsyncView[AsyncModuleType]):
        self.view = view
        self.current_index = 0
        self.queue: list[AsyncModuleType] = []

    def __aiter__(self) -> AsyncIterator[AsyncModuleType]:
        return self

    async def __anext__(self) -> AsyncModuleType:
        if len(self.queue) == 0:
            # Note: this implementation definitely isn't perfect yet and does introduce
            # some rather ugly database strain (since Sugar probably doesn't batch these
            # in some smart manner and just sends of 6 requests). But it's still only
            # one HTTP request, so it's better than fetching records individually.
            new_records = await self.view._module.get_client().bulk(
                self.view[self.current_index],
                self.view[self.current_index + 1],
                self.view[self.current_index + 2],
                self.view[self.current_index + 3],
                self.view[self.current_index + 4],
                self.view[self.current_index + 5],
            )
            self.queue.extend(new_records)
            self.current_index += 6
        record = self.queue.pop(0)
        if record is None:
            raise StopAsyncIteration
        else:
            return record


class AsyncView(
    Generic[AsyncModuleType],
    View[
        AsyncModuleType,
        Awaitable[AsyncModuleType],
        Awaitable[Optional[AsyncModuleType]],
    ],
):
    def __aiter__(self) -> AsyncIterator[AsyncModuleType]:
        return AsyncViewIterator(self)

    def __reversed__(self) -> AsyncIterator[AsyncModuleType]:
        return AsyncViewIterator(self[::-1])

    async def length(self) -> int:
        await self._calculate_range()
        assert isinstance(self._range, range)
        return len(self._range)

    # The following two methods are exactly the same as the the corresponding SyncView
    # implementations, with the exception of them awaiting asynchronous methods.

    async def _fetch_size(self) -> None:
        """Fetch the total number of results that can be queried from this view's
        filters."""
        data = await self._module.get_client().request(
            "get",
            # https://support.sugarcrm.com/Documentation/Sugar_Developer/Sugar_Developer_Guide_11.1/Integration/Web_Services/REST_API/Endpoints/modulecount_GET/
            f"{self._base_endpoint}/count",
            params=self._query_params,
        )
        record_count = data.get("record_count", None)
        if not isinstance(record_count, int):
            raise InvalidSugarResponseError(
                f"got invalid record count: {record_count!r}"
            )
        self._size = record_count

    async def _calculate_range(self) -> None:
        """Apply all pending slicing operations on the cached range object."""
        # This is a pretty messy hack for when this method gets called inside a bulk()
        # session multiple times. The problem is that when self._pending_slices gets
        # emptied further down, only the first instance of this method will get the
        # actual list of pending slices and the other ones will get an empty list -
        # since it was already emptied by the first list. The clean solution here would
        # be to properly synchronize this call so that only one is run at once (because
        # they all return the same thing anyway, and therefore we only need one request
        # ). The problem is that bulk() then thinks that they are all waiting on some
        # external awaitable when instead they are waiting on each other. Since the bulk
        # query is only started when all actions either resolve or are waiting for a
        # (sugar) request, it would end up in a deadlock. To get around this problem,
        # we copy the list of pending slices, so this method can be called in parallel
        # without a problem (let's assume here that the server will return the same
        # record count every time).
        pending_slices = self._pending_slices[:]

        if self._range is None:
            if self._size is None:
                await self._fetch_size()
            assert isinstance(self._size, int)
            self._range = range(0, self._size, 1)

        while len(pending_slices) > 0:
            self._range = self._range[pending_slices.pop(0)]
        self._pending_slices.clear()

    ###################
    # Record fetching #
    ###################

    async def get_by_id(self, key: str) -> AsyncModuleType:
        try:
            return await self._prepare_get_by_id(key)[0]
        except IndexError:
            pass
        raise KeyError(key)

    async def _get_by_offset(self, offset: int) -> Optional[AsyncModuleType]:
        preparation = self._prepare_get_by_offset(offset)
        if isinstance(preparation, tuple):
            method, endpoint, params = preparation
            data = await self._module.get_client().request(
                method, endpoint, params=params
            )
            return self._finalize_get_by_offset(offset, data)
        else:
            return preparation

    async def get_by_index(self, index: int) -> AsyncModuleType:
        await self._calculate_range()
        offset = self._index_to_offset(index)
        if offset is None:
            raise IndexError(index)
        record = await self._get_by_offset(offset)
        if record is None:
            raise IndexError(index)
        return record



from __future__ import annotations

import abc
from numbers import Number
from typing import Awaitable  # noqa: F401
from typing import TYPE_CHECKING, Any, Generic, Optional, Type, TypeVar, Union, overload

from zucker.filtering import NegatableFilter, NullishFilter, NumericFilter, ValuesFilter
from zucker.utils import ApiType, JsonType

if TYPE_CHECKING:
    from zucker.model.module import AsyncModule, BaseModule, SyncModule, UnboundModule


NativeType = TypeVar("NativeType")
SyncGetType = TypeVar("SyncGetType")
AsyncGetType = TypeVar("AsyncGetType")
SetType = TypeVar("SetType")
Self = TypeVar("Self", bound="Field[Any, Awaitable[Any]]")


class Field(Generic[SyncGetType, AsyncGetType], abc.ABC):
    """Base class for all fields.

    This class accepts type variables which resemble the native type that is returned
    when the field accessed, for example as ``record.some_value`` where ``some_value``
    is the field instance. The generic arguments give the return type of that call,
    depending on whether the field is placed in a synchronous or an asynchronous
    module.
    """

    def __init__(self, api_name: Optional[str] = None, **kwargs: Any):
        self._api_name = api_name
        self._name: Optional[str] = "test"
        # Check the name to make sure it's valid.
        self.name  # noqa
        self._name = None

    def __set_name__(self, owner: BaseModule, name: str) -> None:
        self._name = name
        # Check the name to make sure it's valid.
        self.name  # noqa

    @overload
    def __get__(
        self: Self, instance: SyncModule, owner: Type[BaseModule]
    ) -> SyncGetType:
        ...

    @overload
    def __get__(
        self: Self, instance: AsyncModule, owner: Type[BaseModule]
    ) -> AsyncGetType:
        ...

    @overload
    def __get__(
        self: Self, instance: UnboundModule, owner: Type[BaseModule]
    ) -> SyncGetType:
        ...

    @overload
    def __get__(self: Self, instance: None, owner: Type[BaseModule]) -> Self:
        ...

    def __get__(
        self: Self,
        instance: Union[SyncModule, AsyncModule, UnboundModule, None],
        owner: Type[BaseModule],
    ) -> Union[SyncGetType, AsyncGetType, Self]:
        from ..module import BaseModule

        if isinstance(instance, BaseModule):
            # Here, the field is accessed as a property on a record, like this:
            #   the_first_name = record.first_name
            # In this case, the actual field type determines which data is returned.
            value = self._get_value(instance)
            # This type is ignored because because MyPy cannot parse generic self
            # argument types and just flat-out ignores them currently. Hence,
            # self._get_value() is actually an Any (since self is a 'Self' which is
            # Field[Any] or some subtype).
            # See also: https://github.com/python/mypy/issues/2354
            return value  # type:ignore
        elif instance is None:
            # The other case is when the field gets referenced directly on the class,
            # for example when building queries:
            #   module.filter(Lead.name == "Apple")
            # Here we return the field itself again, because then we can build filters
            # and use other APIs from the class.
            return self

        else:
            raise AttributeError(
                f"Field() objects should only created inside bound modules - got "
                f"{instance!r}"
            )

    @property
    def name(self) -> str:
        if self._api_name is not None:
            result = self._api_name
        elif self._name is None:
            raise RuntimeError(
                "Could not retrieve the field's model name. Check for the correct "
                "Field() usage - otherwise this is a bug."
            )
        else:
            result = self._name

        if not isinstance(result, str):
            raise TypeError(f"field name must be a string, got {result!r}")
        if result == "" or " " in result:
            raise ValueError("field name may not be empty and must not contain spaces")

        return result

    @overload
    def _get_value(self, record: SyncModule) -> SyncGetType:
        ...

    @overload
    def _get_value(self, record: AsyncModule) -> AsyncGetType:
        ...

    @overload
    def _get_value(self, record: UnboundModule) -> SyncGetType:
        ...

    @abc.abstractmethod
    def _get_value(
        self, record: Union[SyncModule, AsyncModule, UnboundModule]
    ) -> Union[SyncGetType, AsyncGetType]:
        ...


class MutableField(
    Generic[SyncGetType, AsyncGetType, SetType],
    Field[SyncGetType, AsyncGetType],
    abc.ABC,
):
    """Base class for fields that are mutable.

    Being "mutable" means that users can not only read values, but also write (and save)
    new values for the field. Since setting could potentially take a different type,
    there are two generic arguments for this field. For example, a date field could
    return native date objects but additionally accept strings for setting.
    """

    def __set__(self, instance: BaseModule, value: SetType) -> None:
        from ..module import BaseModule

        if not isinstance(instance, BaseModule):
            raise AttributeError
        self._set_value(instance, value)

    @abc.abstractmethod
    def _set_value(self, record: BaseModule, value: SetType) -> None:
        ...


class ScalarField(Generic[NativeType, ApiType], Field[NativeType, NativeType], abc.ABC):
    """Scalar fields are fields that also support some basic filtering operations."""

    ##################################
    # Getting / setting field values #
    ##################################

    def _get_value(self, record: BaseModule) -> NativeType:
        raw_value = record.get_data(self.name)

        if raw_value is None:
            raise AttributeError(
                f"Trying to access an undefined field {self.name!r} in record "
                f"{type(record)!r}. Either add the field to the module "
                f"definition or check for the correct spelling."
            )
        return self.load_value(raw_value)

    @classmethod
    @abc.abstractmethod
    def load_value(cls, raw_value: JsonType) -> NativeType:
        """Load a value from the API into a native data type.

        :param raw_value: Response from the API for this field. This will be a JSON
            primitive, which should either be returned as-is (where appropriate) or
            converted into a native Python data type.
        :returns: A Python data type for this field.
        """

    @classmethod
    @abc.abstractmethod
    def serialize(cls, value: Union[NativeType, ApiType]) -> ApiType:
        """Serialize a native data type into something the API can take back for
        saving.

        Note: this method also supports "serializing" api types, which generally should
        just verify their validity and return them as-is.
        """

    ###################
    # Filter building #
    ###################

    def __eq__(  # type: ignore[override]
        self, other: Optional[Union[NativeType, ApiType]]
    ) -> NegatableFilter[Any]:
        """Filter for exact values of this field.

        Depending on type of the given value, this is is equivalent to one of the other
        filtering methods:

            >>> Person.name == "Ben" # Is the same as Person.name.values("Ben")
            >>> Person.age == 3 # Is the same as Person.age.values(3)
            >>> Person.supervisor == None # Is the same as Person.supervisor.null()
        """
        if other is None:
            return self.null()
        else:
            return self.values(other)

    def __ne__(  # type: ignore[override]
        self, other: Optional[Union[NativeType, ApiType]]
    ) -> NegatableFilter[Any]:
        """Inverse of the ``==`` filter operator.

        Use the ``!=`` operator to exclude specific values:

            >>> Person.name != "Ben" # Is the same as ~(Person.name.values("Ben"))
            >>> Person.supervisor != None # Is the same as ~(Person.supervisor.null())
        """
        return ~(self.__eq__(other))

    def values(self, *values: Union[NativeType, ApiType]) -> ValuesFilter[ApiType]:
        """Filter for exact values of this field.

        Most basic use for this filter is finding objects by value. The filter

            >>> Person.name.values("Ben")

        will return objects who's name is 'Ben'.

        This filter takes one or more arguments. It matches entries where this set
        directly contains the field's value, for example:

            >>> Person.name.values("Paul", "Spencer")

        will match objects who's name is either 'Paul' or 'Spencer'.

        Inverting this filter yields a 'not-equal' filter, for example:

            >>> ~Person.name.values("Mike")

        This query will match all objects where the name is not equal to 'Mike'.

        The above examples are also available as a shorthand through the equals
        operator (although you can only check for a single value here):

            >>> Person.name == "Ben"
            >>> Person.name != "Ben"
        """
        return ValuesFilter(
            self.name, *tuple(self.serialize(raw_value) for raw_value in values)
        )

    def null(self) -> NullishFilter:
        """Filter for whether the field is null.

        Use the filter like this:

            >>> Person.employer.null()

        This will return objects where the 'employer' field is not set.

        To find only objects where a field is explicitly not null, invert the filter:

            >>> ~Person.employer.null()

        As a shorthand, you can also use the equals operator for the above examples:

            >>> Person.employer == None
            >>> Person.employer != None
        """
        return NullishFilter(self.name)


class MutableScalarField(
    Generic[NativeType, ApiType],
    ScalarField[NativeType, ApiType],
    # Setting is supported by both the API type and the native type.
    MutableField[NativeType, NativeType, Union[NativeType, ApiType]],
    abc.ABC,
):
    def _set_value(self, record: BaseModule, value: Union[ApiType, NativeType]) -> None:
        raw_value = self.serialize(value)
        record._updated_data[self.name] = raw_value


class MutableNumericField(
    Generic[ApiType], MutableScalarField[ApiType, ApiType], abc.ABC
):
    def __lt__(self, other: Number) -> NumericFilter:
        """Filter for values less than the specified scalar.

        :param other: Only records with a field value less than this value will be
            included.
        """
        return NumericFilter(self.name, other, greater=False, equal=False)

    def __lte__(self, other: Number) -> NumericFilter:
        """Filter for values less than or equal to the specified scalar.

        :param other: Only records with a field value less than or equal to this value
            will be included.
        """
        return NumericFilter(self.name, other, greater=False, equal=True)

    def __gt__(self, other: Number) -> NumericFilter:
        """Filter for values less than the specified scalar.

        :param other: Only records with a field value greater than this value will be
            included.
        """
        return NumericFilter(self.name, other, greater=True, equal=False)

    def __gte__(self, other: Number) -> NumericFilter:
        """Filter for values greater than or equal to the specified scalar.

        :param other: Only records with a field value greater than or equal to this
            value will be included.
        """
        return NumericFilter(self.name, other, greater=True, equal=True)



from __future__ import annotations

from abc import ABC
from numbers import Number
from typing import Any, Generic, Literal, Sequence, TypeVar, Union, cast

from ..utils import ApiType, JsonMapping
from .combining import FilterSet
from .types import Combinator

Filters = TypeVar("Filters")


class BasicFilter(Generic[Filters], ABC):
    """This is the most basic sort of filter - it checks if some specific field matches
    a condition.

    Subclasses of this type determine the exact condition that is tested (and how that
    condition is mapped to Sugar's API syntax). These filters can be combined with the
    ``&`` and ``|`` operators to create more complex filter sets.
    """

    def __init__(self, field_name: str, filters: Filters):
        self.field_name = field_name
        self.filters = filters

    def __or__(self, other: BasicFilter[Any]) -> FilterSet:
        if not isinstance(other, BasicFilter):
            return NotImplemented  # type: ignore
        return FilterSet(Combinator.OR, self, other)

    def __and__(self, other: BasicFilter[Any]) -> FilterSet:
        if not isinstance(other, BasicFilter):
            return NotImplemented  # type: ignore
        return FilterSet(Combinator.AND, self, other)

    @property
    def operator(self) -> str:
        raise NotImplementedError(
            "subclasses of BasicFilter must implement the operator property"
        )

    @property
    def filter_value(self) -> Filters:
        return self.filters

    def build_filter(self) -> JsonMapping:
        return {self.field_name: {self.operator: self.filter_value}}


class NegatableFilter(Generic[Filters], BasicFilter[Filters], ABC):
    def __invert__(self) -> NegatableFilter[Filters]:
        raise NotImplementedError(
            "subclasses of NegatableFilter must implement the __invert__ protocol"
        )


class ValuesFilter(
    Generic[ApiType], NegatableFilter[Union[ApiType, Sequence[ApiType]]]
):
    """Basic filter that checks if the field exactly matches specified values.

    The constructor takes a field name and at least one value to test for. When one
    value is given, an equality check (using ``$equals`` in the Sugar API) will be
    produced. If more then one value is provided, it will switch to the ``$in``
    operator, which allows checking for multiple accepted values.

    This filter can be negated using the ``~`` operator.
    """

    def __init__(self, field_name: str, *filters: ApiType):
        if any(not isinstance(item, str) for item in filters):
            # TODO This doesn't currently match the type defintion above.
            raise TypeError("values for a value filter must strings")
        if len(filters) == 0:
            raise ValueError("did not provide any values for a value filter")

        self.negated = False
        self._single = False

        if len(filters) == 1:
            super().__init__(field_name, filters[0])
            self._single = True
        else:
            super().__init__(field_name, filters)

    def __invert__(self) -> ValuesFilter[ApiType]:
        result: ValuesFilter[ApiType]
        if self._single:
            result = ValuesFilter(self.field_name, cast("ApiType", self.filters))
        else:
            result = ValuesFilter(
                self.field_name, *(cast("Sequence[ApiType]", self.filters))
            )
        result.negated = not self.negated
        return result

    @property
    def operator(self) -> str:
        if self._single:
            return "$not_equals" if self.negated else "$equals"
        else:
            return "$not_in" if self.negated else "$in"


class NullishFilter(NegatableFilter[None]):
    """Basic filter that checks if the field is None (or ``null`` in Sugar).

    This filter can be negated using the ``~`` operator.
    """

    def __init__(self, field_name: str):
        super().__init__(field_name, None)

        self.negated = False

    def __invert__(self) -> NullishFilter:
        result = NullishFilter(self.field_name)
        result.negated = not self.negated
        return result

    @property
    def operator(self) -> str:
        return "$not_null" if self.negated else "$is_null"


class StringFilter(BasicFilter[str], ABC):
    def __init__(self, field_name: str, value: str):
        if not isinstance(value, str):
            raise TypeError(
                f"string filters are only applicable for strings, got {type(value)!r}"
            )
        if len(value) == 0:
            raise ValueError("cannot filter for empty strings")

        super().__init__(field_name, value)


class StringStartsFilter(StringFilter):
    """Basic filter that checks if the string field starts with a given pattern."""

    @property
    def operator(self) -> str:
        return "$starts"


class StringEndsFilter(StringFilter):
    """Basic filter that checks if the string field ends with a given pattern."""

    @property
    def operator(self) -> str:
        return "$ends"


class StringContainsFilter(StringFilter):
    """Basic filter that checks if the string field contains a given pattern."""

    @property
    def operator(self) -> str:
        return "$ends"


class NotEmptyFilter(BasicFilter[Literal[""]]):
    def __init__(self, field_name: str):
        super().__init__(field_name, "")

    @property
    def operator(self) -> str:
        return "$not_empty"


class NumericFilter(NegatableFilter[Number]):
    def __init__(
        self,
        field_name: str,
        value: Number,
        *,
        greater: bool,
        equal: bool,
    ):
        if not isinstance(value, Number):
            raise TypeError(
                f"numeric filters only work with numbers, got {type(value)!r}"
            )

        super().__init__(field_name, value)
        self._greater = greater
        self._equal = equal

    def __invert__(self) -> NumericFilter:
        # When inverting,
        #   >  becomes <=
        #   >= becomes <
        #   <  becomes >=
        #   <= becomes >
        return NumericFilter(
            self.field_name,
            self.filters,
            greater=not self._greater,
            equal=not self._equal,
        )

    @property
    def operator(self) -> str:
        # Choose one of $gt, $gte, $lt and $lte.
        return f"${'gt' if self._greater else 'lt'}{'e' if self._equal else ''}"



from typing import Sequence, Union

import colored

from zucker.client import SyncClient
from zucker.model import fields

from ..inspection import InspectedField, InspectedModule, inspect_modules_with_fields

TERM = colored.attr("reset")
TERM_MODULE_NAME = colored.fg("blue")
TERM_MODULE_TITLE = TERM_MODULE_NAME + colored.attr("bold")
TERM_FIELD_NAME = colored.fg("red")
TERM_ARGUMENT_NAME = colored.fg("light_gray")


def indent(lines: Union[str, Sequence[str]], steps: int = 1) -> str:
    all_lines: Sequence[str]
    if isinstance(lines, str):
        all_lines = lines.split("\n")
    elif isinstance(lines, Sequence):
        all_lines = "\n".join(lines).split("\n")
    else:
        raise TypeError
    indentation = "  " * steps
    return indentation + f"\n{indentation}".join(all_lines)


def format_field(field: InspectedField) -> str:
    formatted_arguments = []
    for name, value in field.arguments.items():
        # Skip the api_name argument if it matches the field name (because then it's
        # redundant).
        if name == "api_name" and value == field.name:
            continue

        formatted_value: str
        if isinstance(value, InspectedModule):
            formatted_value = f"{TERM_MODULE_NAME}{value.name}{TERM}"
        else:
            formatted_value = repr(value)
        formatted_arguments.append(
            f"{TERM_ARGUMENT_NAME}{name}={TERM}{formatted_value}"
        )

    return " ".join(
        [
            f"{TERM_FIELD_NAME}{field.name}{TERM}",
            f"({field.field_type.__name__})",
            *formatted_arguments,
        ]
    )


def format_module(module: InspectedModule) -> str:
    argument_lines = [
        f"{key} = {value!r}"
        for key, value in module.class_arguments.items()
        # Skip the api_name argument if it matches the module name (because then it's
        # redundant).
        if key != "api_name" or value != module.name
    ]
    field_lines = [
        format_field(field)
        for field in sorted(module.fields, key=lambda field: field.name)
    ]

    return (
        "\n".join(
            [f"{TERM_MODULE_NAME}{module.name}{TERM}"]
            + (
                [
                    indent("Arguments:", 1),
                    indent(argument_lines, 2),
                ]
                if len(argument_lines) > 0
                else []
            )
            + ([indent(field_lines, 1)] if len(field_lines) > 0 else [])
        )
        + "\n"
    )


def run_inspect(client: SyncClient, **kwargs: None) -> None:
    client.fetch_metadata("modules")
    modules_metadata = client.get_metadata_item("modules")
    modules = inspect_modules_with_fields(modules_metadata)

    for module in sorted(modules, key=lambda module: module.name):
        print(format_module(module))



from enum import Enum
from typing import Protocol, runtime_checkable

from ..utils import JsonMapping


@runtime_checkable
class GenericFilter(Protocol):
    def build_filter(self) -> JsonMapping:
        ...


class Combinator(Enum):
    OR = "$or"
    AND = "$and"



import argparse
import sys
from getpass import getpass
from typing import Optional, Sequence, Union

from zucker.client import RequestsClient

from .inspect import run_inspect


class PromptPasswordAction(argparse.Action):
    # argparse.Action has a bunch of keyword arguments that we don't bother typing here.
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        kwargs.setdefault("nargs", 0)
        super().__init__(*args, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Union[str, Sequence[str], None],
        option_string: Optional[str] = None,
    ) -> None:
        if option_string in self.option_strings:
            if getattr(namespace, self.dest, None) is not None:
                parser.error(
                    "plaintext and prompt password options are mutually exclusive"
                )
            setattr(namespace, self.dest, getpass())


def run_from_command_line() -> None:
    parser = argparse.ArgumentParser(
        description="Automatically generate zucker model classes from a remote Sugar "
        "CRM instance.",
    )
    subparsers = parser.add_subparsers(title="subcommands")

    parser.add_argument(
        "-b",
        "--base-url",
        dest="base_url",
        metavar="URL",
        required=True,
        help="base URL of the Sugar CRM server",
    )
    parser.add_argument(
        "-c",
        "--client-platform",
        dest="client_platform",
        metavar="PLATFORM",
        default="zucker",
        help="client platform on the server",
    )
    parser.add_argument(
        "-u",
        "--username",
        dest="username",
        help="username to use for authentication",
    )
    parser.add_argument(
        "--plaintext-password",
        dest="password",
        help="plaintext password to use for authentication (consider using -P instead)",
    )
    parser.add_argument(
        "-P",
        "--prompt-password",
        dest="password",
        action=PromptPasswordAction,
        help="prompt for authentication password",
    )
    parser.add_argument(
        "--disable-ssl-verification",
        dest="verify_ssl",
        action="store_false",
        default=True,
        help="disable SSL certificate verification",
    )

    inspect_parser = subparsers.add_parser(
        "inspect",
        description="Build a list of the server's data model. This can be used to "
        "manually find the relevant fields.",
        help="extract the data model for manual inspection",
    )
    inspect_parser.set_defaults(callable=run_inspect)

    args = parser.parse_args()

    if getattr(args, "password", None) is None or len(args.password) == 0:
        parser.error("no authentication password was provided")
    if getattr(args, "callable", None) is None:
        parser.print_help()
        sys.exit(2)

    client = RequestsClient(
        **{
            key: getattr(args, key)
            for key in (
                "base_url",
                "username",
                "password",
                "client_platform",
                "verify_ssl",
            )
        }
    )
    args.callable(client, **args.__dict__)



from __future__ import annotations

import abc
import asyncio
import urllib.parse
from datetime import datetime, timedelta
from json import dumps as dump_json
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Iterator,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from zucker.exceptions import (
    InvalidSugarResponseError,
    SugarError,
    UnfetchedMetadataError,
    ZuckerException,
)
from zucker.utils import JsonMapping, JsonType, MutableJsonMapping, check_json_mapping

if TYPE_CHECKING:
    from zucker.model.module import AsyncModule, BoundModule, SyncModule  # noqa: F401

_T = TypeVar("_T")
_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")
_T5 = TypeVar("_T5")
_T6 = TypeVar("_T6")


class BaseClient(abc.ABC):
    """Connection handler that handles communicating with a SugarCRM instance.

    This is the main entry point for interfacing with models. After creating a new
    instance, the server's modules are available as attributes on the client. Concrete
    client implementations (for different transport mechanisms) should implement one of
    :class:`~SyncClient` or :class:`~AsyncClient`.
    """

    def __init__(
        self,
        base_url: str,
        username: str,
        password: str,
        *,
        client_platform: str = "zucker",
        verify_ssl: bool = True,
    ):
        """Construct a client instance.

        This may take a few seconds, as it performs initial authentication as well as querying
        the server for metadata about it's configuration.

        :param base_url: The URL of the SugarCRM installation to connect to.
        :param username: Username to authenticate with.
        :param password: Password to authenticate with.
        :param client_platform: OAuth platform string.
        :param verify_ssl: Set this to false to disable verification of the server's SSL
            certificate. This should only be used while testing!
        """
        values = (base_url, username, password, client_platform)
        if any(not isinstance(value, str) for value in values):
            raise TypeError(
                "all relevant parameters must be provided to create a Sugar client"
            )
        if any(len(value) == 0 for value in values):
            raise ValueError(
                "all relevant parameters must be provided to create a Sugar client"
            )

        self.base_url = base_url

        # The authentication is stored as one of two tuple types, depending on the
        # current state:
        self._authentication: Union[
            # This is the initial configuration, holding the username and password.
            Tuple[Literal[False], str, str],
            # After the initial OAuth step has completed, this tuple is stored instead.
            # It contains the access token, the refresh token and an expiry timestamp,
            # in that order.
            Tuple[Literal[True], str, str, datetime],
        ] = (False, username, password)
        self._client_platform = client_platform
        self._verify_ssl = verify_ssl

        self._metadata: MutableJsonMapping = {}

    def __contains__(self, item: Union[str, Type[BoundModule[Any]]]) -> bool:
        """Check if the server supports a given module name."""
        from zucker.model.module import BoundModule

        if isinstance(item, type) and issubclass(item, BoundModule):
            return item._api_name in self.module_names
        elif isinstance(item, str):
            return item in self.module_names
        else:
            raise TypeError(
                f"contains checks on a SugarClient are only supported with strings and "
                f"Module classes, got {type(item)!r}"
            )

    @abc.abstractmethod
    def fetch_metadata(self, *types: str) -> Union[None, Awaitable[None]]:
        """Make sure server metadata for the given set of types is available."""

    def get_metadata_item(self, type_name: str) -> JsonMapping:
        """Return the cached value of a given metadata item.

        :param type_name: Name of the metadata's type, as defined by Sugar.
        :raise UnfetchedMetadataError: This error will be raised when the corresponding
            metadata has not been fetched yet. In that case, call :meth:`fetch_metadata`
            to populate the cache.
        """
        if type_name not in self._metadata:
            raise UnfetchedMetadataError(
                f"metadata field {type_name!r} is not available"
            )
        try:
            return check_json_mapping(self._metadata[type_name])
        except TypeError:
            raise InvalidSugarResponseError("got invalid server metadata")

    @property
    def module_names(self) -> Iterator[str]:
        """List of all modules that are available.

        This requires fetching the ``full_module_list`` metadata item.
        """
        full_module_list = self.get_metadata_item("full_module_list")
        assert isinstance(full_module_list, Mapping)

        yield from (
            name for name in full_module_list.keys() if not name.startswith("_")
        )

    @property
    def server_info(self) -> Tuple[str, str, str]:
        """Sever information tuple.

        This requires fetching the ``server_info`` metadata item.

        :return: A 3-tuple that follows the syntax (``flavor``, ``version``, ``build``).
        """
        server_info = self.get_metadata_item("server_info")
        assert isinstance(server_info, Mapping)

        flavor = server_info["flavor"]
        assert isinstance(flavor, str)

        version = server_info["version"]
        assert isinstance(version, str)

        build = server_info["build"]
        assert isinstance(build, str)

        return flavor, version, build

    @property
    def authenticated(self) -> bool:
        """Shows whether initial authentication with the server has already been
        performed.
        """
        return self._authentication[0]

    def _prepare_authentication(
        self,
    ) -> Optional[Tuple[str, str, Mapping[str, str]]]:
        """Build the payload currently required to process authentication.

        The output of this function will be one of:

        - ``None``, when authentication information is still valid
        - ``("initial authentication", str, dict)`` for the first authentication with
          username and password
        - ``("authentication token renewal", str, dict)`` for renewing an active token

        For the second and third outputs, the tuple contains the name of the action
        being performed (for error logging), the API endpoint and the ``data`` parameter
        that should be used for performing the authentication request.
        """
        if self._authentication[0] is True:
            # Initial OAuth has already happened. The token will be renewed.
            _, _, refresh_token, expire_timestamp = self._authentication
            if expire_timestamp > datetime.now() + timedelta(minutes=10):
                return None
            return (
                "authentication token renewal",
                "oauth2/token/",
                {
                    "grant_type": "refresh_token",
                    "client_id": "sugar",
                    "client_secret": "",
                    "refresh_token": refresh_token,
                    "platform": self._client_platform,
                },
            )
        else:
            # No initial authentication has happened yet. Retrieve the initial tokens.
            _, username, password = self._authentication
            return (
                "initial authentication",
                "oauth2/token/",
                {
                    "grant_type": "password",
                    "client_id": "sugar",
                    "client_secret": "",
                    "username": username,
                    "password": password,
                    "platform": self._client_platform,
                },
            )

    def _finalize_authentication(
        self, auth_job_name: str, response_code: int, data: Any
    ) -> None:
        """Process the result from the authentication call and save the required
        tokens."""
        if not (200 <= response_code < 300):
            raise ZuckerException(
                f"{auth_job_name} failed with status code {response_code}"
            )

        response_json = check_json_mapping(data)

        if (
            "access_token" not in response_json
            or "refresh_token" not in response_json
            or "expires_in" not in response_json
        ):
            raise InvalidSugarResponseError(
                "missing response_json fields from authentication result"
            )

        access_token = response_json["access_token"]
        refresh_token = response_json["refresh_token"]
        if not isinstance(access_token, str) or not isinstance(refresh_token, str):
            raise InvalidSugarResponseError(
                f"bad authentication data: expected token strings, got "
                f"{type(access_token)!r} and {type(refresh_token)!r}"
            )

        expires_in = response_json["expires_in"]
        if not isinstance(expires_in, (int, float)):
            raise InvalidSugarResponseError(
                f"bad authentication data: expected exipry timestamp as a number, got "
                f"{type(expires_in)!r}"
            )
        expire_timestamp = datetime.now() + timedelta(seconds=expires_in)

        self._authentication = (True, access_token, refresh_token, expire_timestamp)

    def _finalize_request(self, response_code: int, data: Any) -> JsonMapping:
        """Process the response from an API request and return a type-checked
        ``JsonMapping`` type.
        """
        try:
            response_json = check_json_mapping(data)
        except TypeError:
            raise InvalidSugarResponseError("got invalid JSON response from Sugar")
        if 200 <= response_code < 300:
            return response_json
        else:
            raise SugarError(response_code, response_json)

    @abc.abstractmethod
    def close(self) -> Union[None, Awaitable[None]]:
        """Perform cleanup operations.

        This method should be called when the client is no longer used. Further
        operations on the client or any of its modules are consider undefined behavior
        afterwards. Some clients may be able to reinitialize transparently after
        closing while others may be unusable.
        """

    @abc.abstractmethod
    def request(
        self,
        method: str,
        endpoint: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        data: Optional[JsonMapping] = None,
        json: Optional[JsonMapping] = None,
    ) -> Union[JsonMapping, Awaitable[JsonMapping]]:
        """Handle a request to the CRM's REST API.

        This will make sure that authentication is set up and the correct URL is built.
        Parameters are akin to those defined by popular HTTP libraries (although not all
        are supported).

        :param method: HTTP verb to use.
        :param endpoint: Desired endpoint name (the part after `/rest/v??_?/`).
        :param params: Dictionary of query parameters to add to the URL.
        :param data: Request form data.
        :param json: Request body that will be JSON-encoded. This is mutually exclusive
            with ``data``.
        """


class SyncClient(BaseClient, abc.ABC):
    def close(self) -> None:
        pass

    @abc.abstractmethod
    def raw_request(
        self,
        method: str,
        endpoint: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        data: Optional[JsonMapping] = None,
        json: Optional[JsonMapping] = None,
    ) -> tuple[int, JsonMapping]:
        """Request handling method that should be overridden by client implementations.

        This takes the same parameters as :meth:`BaseClient.request`.

        :returns: A tuple containing the response's status code and the JSON object
            gotten from the API.
        """

    def request(
        self,
        method: str,
        endpoint: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        data: Optional[JsonMapping] = None,
        json: Optional[JsonMapping] = None,
    ) -> JsonMapping:
        auth_payload = self._prepare_authentication()

        if auth_payload is not None:
            auth_job_name, auth_endpoint, auth_data = auth_payload
            response_code, response_json = self.raw_request(
                "post",
                auth_endpoint,
                data=auth_data,
            )
            self._finalize_authentication(auth_job_name, response_code, response_json)

        response_code, response_json = self.raw_request(
            method, endpoint, params=params, data=data, json=json
        )
        return self._finalize_request(response_code, response_json)

    def fetch_metadata(self, *types: str) -> None:
        """Make sure server metadata for the given set of types is available."""
        self._metadata.update(
            self.request(
                "get",
                "metadata",
                params={"type_filter": ",".join(types)},
            )
        )


class AsyncClient(BaseClient):
    def __init__(
        self,
        base_url: str,
        username: str,
        password: str,
        *,
        client_platform: str = "zucker",
        verify_ssl: bool = True,
    ):
        super().__init__(
            base_url,
            username,
            password,
            client_platform=client_platform,
            verify_ssl=verify_ssl,
        )

        self._handle_bulk: Optional[
            Callable[[JsonMapping], Awaitable[tuple[int, JsonMapping]]]
        ] = None

    async def close(self) -> None:
        pass

    @abc.abstractmethod
    async def raw_request(
        self,
        method: str,
        endpoint: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        data: Optional[JsonMapping] = None,
        json: Optional[JsonMapping] = None,
    ) -> tuple[int, JsonType]:
        """Request handling method that should be overridden by client implementations.

        This takes the same parameters as :meth:`BaseClient.request`.

        :returns: A tuple containing the response's status code and the JSON object
            gotten from the API.
        """

    async def _ensure_authentication(self) -> None:
        auth_payload = self._prepare_authentication()

        if auth_payload is not None:
            auth_job_name, auth_endpoint, auth_data = auth_payload
            response_code, response_json = await self.raw_request(
                "post",
                auth_endpoint,
                data=auth_data,
            )
            self._finalize_authentication(auth_job_name, response_code, response_json)

    async def request(
        self,
        method: str,
        endpoint: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        data: Optional[JsonMapping] = None,
        json: Optional[JsonMapping] = None,
        allow_bulk: bool = True,
    ) -> JsonMapping:
        response_code: int
        response_json: JsonMapping

        if self._handle_bulk is not None and allow_bulk:
            # Important: when making changes, make sure to not await anything in this
            # branch except the final self._handle bulk so we don't get any race
            # conditions (we don't assume thread-safety, however).

            if data is not None:
                raise RuntimeError(
                    "data argument for requests is not supported in bulk contexts"
                )

            url = f"/v11_5/{endpoint}"
            if params is not None:
                url += "?"
                url += "&".join(
                    f"{urllib.parse.quote_plus(key)}={urllib.parse.quote_plus(value)}"
                    for key, value in params.items()
                )
            request_definition = dict(url=url, method=method.upper())
            if json is not None:
                request_definition["data"] = dump_json(json)

            response_code, response_json = await self._handle_bulk(request_definition)
        else:
            await self._ensure_authentication()

            response_code, response_raw_json = await self.raw_request(
                method, endpoint, params=params, data=data, json=json
            )
            response_json = check_json_mapping(response_raw_json)

        return self._finalize_request(response_code, response_json)

    # Typing for bulk() is a bit stupid because the Python typing system doesn't (yet)
    # support unknown-length type variable lists. See here:
    # https://github.com/python/typeshed/pull/1550
    # https://github.com/python/typeshed/blob/master/stdlib/asyncio/tasks.pyi#L44-L47

    @overload
    async def bulk(self, action_1: Awaitable[_T1], /) -> tuple[_T1]:
        ...

    @overload
    async def bulk(
        self, action_1: Awaitable[_T1], action_2: Awaitable[_T2], /
    ) -> tuple[_T1, _T2]:
        ...

    @overload
    async def bulk(
        self,
        action_1: Awaitable[_T1],
        action_2: Awaitable[_T2],
        action_3: Awaitable[_T3],
        /,
    ) -> tuple[_T1, _T2, _T3]:
        ...

    @overload
    async def bulk(
        self,
        action_1: Awaitable[_T1],
        action_2: Awaitable[_T2],
        action_3: Awaitable[_T3],
        action_4: Awaitable[_T4],
        /,
    ) -> tuple[_T1, _T2, _T3, _T4]:
        ...

    @overload
    async def bulk(
        self,
        action_1: Awaitable[_T1],
        action_2: Awaitable[_T2],
        action_3: Awaitable[_T3],
        action_4: Awaitable[_T4],
        action_5: Awaitable[_T5],
        /,
    ) -> tuple[_T1, _T2, _T3, _T4, _T5]:
        ...

    @overload
    async def bulk(
        self,
        action_1: Awaitable[_T1],
        action_2: Awaitable[_T2],
        action_3: Awaitable[_T3],
        action_4: Awaitable[_T4],
        action_5: Awaitable[_T5],
        action_6: Awaitable[_T6],
        /,
    ) -> tuple[_T1, _T2, _T3, _T4, _T5, _T6]:
        ...

    @overload
    async def bulk(self, *actions: Awaitable[Any]) -> tuple[Any, ...]:
        ...

    async def bulk(self, *actions: Awaitable[Any]) -> tuple[Any, ...]:
        """Run a sequence of actions that require server communication together.

        This will use Sugar's `Bulk API`_ to batch all actions together and send them
        as a single HTTP request. It works similarly to :func:`asyncio.gather` in that
        this method will resolve once the result for all provided awaitables is
        available and return them as a tuple. You can also use this method to gather
        together multiple awaitables where only one of them uses the server. If an
        action doesn't resolve after the first request (for example because it needs a
        second one), a second batch will be started.

        Do not use this is threaded environments -- the implementation is not
        thread-safe.

        .. _Bulk API: https://support.sugarcrm.com/Documentation/Sugar_Developer/Sugar_Developer_Guide_11.3/Integration/Web_Services/REST_API/Endpoints/bulk_POST/
        """
        # These dictionaries store the prepared requests that should be sent. Each
        # definition contains a JSON object in the form accepted by the bulk Sugar API
        # (see the link in this method's docstring). Every request that gets started
        # gets a unique key (an int). Next to the definition, we also store an Event
        # that will get set once a response is available. The first dictionary here will
        # hold all requests that are currently waiting to be processed. The second
        # contains responses for those requests where a /bulk call has returned data.
        next_key = 0
        request_definitions = dict[int, tuple[JsonMapping, asyncio.Event]]()
        responses = dict[int, tuple[int, JsonMapping]]()

        # This event will be set whenever the number of waiting requests change.
        counting_event = asyncio.Event()

        async def handle_bulk(
            request_definition: JsonMapping,
        ) -> tuple[int, JsonMapping]:
            """Callback for handling bulk requests. This is called from the actual
            request() method and will queue the request, wait for the corresponding
            batch is done and then return the result."""
            nonlocal next_key

            request_event = asyncio.Event()

            # This shouldn't create any race conditions because we are not targeting
            # thread safety (same goes for the counting stuff further down):
            request_key = next_key
            next_key += 1
            request_definitions[request_key] = request_definition, request_event
            counting_event.set()

            await request_event.wait()

            try:
                return responses.pop(request_key)
            finally:
                counting_event.set()

        self._handle_bulk = handle_bulk
        action_tasks = [asyncio.create_task(action) for action in actions]
        for task in action_tasks:
            task.add_done_callback(lambda *_: counting_event.set())

        # This loop will run as long as at least on action hasn't returned yet and
        # therefore may still be waiting on a request. Further, that action may even
        # perform more requests after that. We handle that case by sending requests
        # off in batches - one in each round of this loop.
        while (
            done_task_count := sum(1 for task in action_tasks if task.done())
        ) != len(actions):
            # Wait until all tasks have either completed (for those that don't actually
            # use the server) or are waiting for the actual server request to start (and
            # have therefore registered themselves in the request_definitions list). In
            # most calls, the latter will probably be the case, but we want to block
            # this method when a completely unrelated awaitable gets passed.
            if len(request_definitions) < len(actions) - done_task_count:
                await counting_event.wait()
                counting_event.clear()
                continue

            # Now, we are in a state where every remaining (as in: not done) task is
            # waiting for a request to complete. That means we can now collect them from
            # the definitions dictionary, run them all through the /bulk API endpoint
            # and split the results back up:

            await self._ensure_authentication()

            request_keys = list(request_definitions.keys())
            request_events: list[asyncio.Event] = [
                request_definitions[key][1] for key in request_keys
            ]
            response_code, response_json = await self.raw_request(
                "post",
                "/bulk",
                json=dict(
                    requests=[request_definitions[key][0] for key in request_keys]
                ),
            )
            request_definitions.clear()
            if not isinstance(response_json, Sequence):
                raise InvalidSugarResponseError(
                    f"expected list from bulk API, got {type(response_json)!r}"
                )
            if not len(response_json) == len(request_keys):
                raise InvalidSugarResponseError(
                    f"expected list of length {len(request_keys)} from bulk API, got "
                    f"length {len(response_json)}"
                )

            for key, item in zip(request_keys, response_json):
                if not isinstance(item, Mapping):
                    raise InvalidSugarResponseError(
                        f"expected dictionary in bulk API list, got {type(item)!r}"
                    )
                if (
                    "contents" not in item
                    or not isinstance(item["contents"], Mapping)
                    or "status" not in item
                    or not isinstance(item["status"], int)
                ):
                    raise InvalidSugarResponseError(
                        "got invalid dictionary layout from bulk API"
                    )
                check_json_mapping(item["contents"])
                responses[key] = (item["status"], item["contents"])

            # When these events are set, handle_bulk() will pick it up and take the
            # correct response out of the 'responses' dictionary:
            for event in request_events:
                event.set()

        return await asyncio.gather(*action_tasks)  # type: ignore

    async def fetch_metadata(self, *types: str) -> None:
        """Make sure server metadata for the given set of types is available."""
        self._metadata.update(
            await self.request(
                "get",
                "metadata",
                params={"type_filter": ",".join(types)},
            )
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, Type, Union, cast, overload

from ...exceptions import WrongClientError, WrongParadigmError
from ..view import AsyncModuleType, AsyncView, SyncModuleType, SyncView
from .base import Field

if TYPE_CHECKING:
    from ..module import AsyncModule, SyncModule, UnboundModule

__all__ = ["RelatedField"]


class RelatedField(
    Generic[SyncModuleType, AsyncModuleType],
    Field[SyncView[SyncModuleType], AsyncView[AsyncModuleType]],
):
    """Field that returns a view on a relationship link.

    :param related_module: The module type on the other side of the relationship.
    :param link_name: Name of the link.

    .. note::
      The related module used to initialize this field must be a bound module with the
      same client as the module the field is attached to. You can't mix synchronous and
      asynchronous models here.
    """

    def __init__(
        self,
        related_module: Union[Type[SyncModuleType], Type[AsyncModuleType]],
        link_name: str,
        **kwargs: Any,
    ):
        from ..module import AsyncModule, BaseModule, SyncModule

        if not issubclass(related_module, (SyncModule, AsyncModule)):
            if issubclass(related_module, BaseModule):  # type: ignore
                raise TypeError(
                    f"related fields must be initialized with a bound module - got the "
                    f"unbound variant {related_module!r}"
                )
            else:
                raise TypeError("a related module must be provided")
        if not isinstance(link_name, str):
            raise TypeError("related link names must be strings")
        link_name = link_name.strip()
        if len(link_name) == 0:
            raise ValueError("related link names must be non-empty")

        super().__init__(**kwargs)
        self._related_module: Union[
            Type[SyncModuleType], Type[AsyncModuleType]
        ] = related_module
        self._link_name = link_name

    @overload
    def _get_value(self, record: SyncModule) -> SyncView[SyncModuleType]:
        ...

    @overload
    def _get_value(self, record: AsyncModule) -> AsyncView[AsyncModuleType]:
        ...

    @overload
    def _get_value(self, record: UnboundModule) -> SyncView[SyncModuleType]:
        ...

    def _get_value(
        self, record: Union[SyncModule, AsyncModule, UnboundModule]
    ) -> Union[SyncView[SyncModuleType], AsyncView[AsyncModuleType]]:
        from ..module import AsyncModule, BoundModule, SyncModule

        if not isinstance(record, BoundModule):
            raise TypeError("felated fields can only be accessed on bound modules")

        key = record.get_data("id")
        if key is None:
            raise ValueError("unable to retrieve key for related lookup")

        if record.get_client() is not self._related_module.get_client():
            raise WrongClientError(
                "Related field was accessed from a different client that the defined "
                "module. Make sure that both the module the field is defined in as "
                "as the referring module are bound to the same client."
            )

        if isinstance(record, SyncModule):
            if not issubclass(self._related_module, SyncModule):
                raise WrongParadigmError(
                    f"Cannot instantiate a non-synchronous record of type "
                    f"{self._related_module!r} from a related field on a synchronous "
                    f"module. If this module has multiple variations on different "
                    f"clients, consider refactoring common parts into a BaseModule "
                    f"subclass and re-implementing related fields with the correct "
                    f"targets in each subclass."
                )
            return SyncView(
                cast(Type[SyncModuleType], self._related_module),
                f"{record._api_name}/{key}/link/{self._link_name}",
            )
        elif isinstance(record, AsyncModule):
            if not issubclass(self._related_module, AsyncModule):
                raise WrongParadigmError(
                    f"Cannot instantiate a non-asynchronous record of type "
                    f"{self._related_module!r} from a related field on an asynchronous "
                    f"module. If this module has multiple variations on different "
                    f"clients, consider refactoring common parts into a BaseModule "
                    f"subclass and re-implementing related fields with the correct "
                    f"targets in each subclass."
                )
            return AsyncView(
                cast(Type[AsyncModuleType], self._related_module),
                f"{record._api_name}/{key}/link/{self._link_name}",
            )
        else:
            raise TypeError(
                f"expecting a synchronous or asynchronous record, got {type(record)!r}"
            )



from __future__ import annotations

import copy
from collections.abc import Mapping
from typing import Union

from ..utils import JsonMapping
from .types import Combinator, GenericFilter

FilterOrMapping = Union[GenericFilter, JsonMapping]


class FilterSet:
    def __init__(
        self,
        combinator: Combinator,
        *given_parts: Union[FilterOrMapping, None],
    ):
        self.combinator = combinator
        self._parts = list(given_parts)

        index = 0
        while index < len(self._parts):
            part = self._parts[index]

            if part is None:
                del self._parts[index]
                continue

            if not isinstance(part, (GenericFilter, Mapping)):
                raise TypeError(
                    f"FilterSet parts must be either dictionaries or generic filter "
                    f"compatible objects (like other FilterSets), got {type(part)!r}"
                )

            # Merge together this filterset and the provided one in these three cases:
            # 1) The current and the new filterset have the same combinator. This will
            #    convert constructs like this:
            #      (A or B or (C or (D or E)))
            #    into something flatter like this:
            #      (A or B or C or D or E)
            # 2) The new filterset has at most one part. In that case, that part (if
            #    available) can be merged into this filterset without changing the
            #    output boolean expression:
            #      (A or (B) or C)
            #    becomes
            #      (A or B or C)
            # 3) The new filterset is the only part we currently have. This is the
            #    reverse condition of the last case. The only difference here is that we
            #    don't keep our own combinator but rather that of the new filterset.
            #    Something like this:
            #      (None or (A and B))
            #    will be extracted to:
            #      (A and B)
            if isinstance(part, FilterSet):
                # This variable checks for the third condition above.
                merge_other = all(
                    other_part is None or other_part is part
                    for other_part in self._parts
                )
                if part.combinator == combinator or len(part._parts) < 2 or merge_other:
                    del self._parts[index]
                    self._parts[index:index] = part._parts
                    if merge_other:
                        self.combinator = part.combinator
                    continue

            # Render everything to an actual filter dictionary so that the whole
            # filterset remains immutable.
            if isinstance(part, GenericFilter):
                part = part.build_filter()

            assert isinstance(part, Mapping)

            self._parts[index] = copy.deepcopy(part)
            index += 1

    def __or__(self, other: FilterOrMapping) -> FilterSet:
        return self._combine(self, other, Combinator.OR)

    def __ror__(self, other: FilterOrMapping) -> FilterSet:
        return self._combine(other, self, Combinator.OR)

    def __and__(self, other: FilterOrMapping) -> FilterSet:
        return self._combine(self, other, Combinator.AND)

    def __rand__(self, other: FilterOrMapping) -> FilterSet:
        return self._combine(other, self, Combinator.AND)

    @staticmethod
    def _combine(
        first: FilterOrMapping, second: FilterOrMapping, combinator: Combinator
    ) -> FilterSet:
        if isinstance(first, (GenericFilter, Mapping)) and isinstance(
            second, (GenericFilter, Mapping)
        ):
            return FilterSet(combinator, first, second)
        return NotImplemented

    def build_filter(self) -> JsonMapping:
        return {
            self.combinator.value: [
                copy.deepcopy(
                    part.build_filter() if isinstance(part, GenericFilter) else part
                )
                for part in self._parts
            ]
        }



from typing import Any

from .utils import JsonMapping


class ZuckerException(Exception):
    """Base class for all errors raised by Zucker."""


class WrongParadigmError(ZuckerException):
    """Raised when the sync / async paradigms clash."""


class WrongClientError(ZuckerException):
    """Raised when an operation receives an unexpected client instance."""


class SugarError(ZuckerException):
    """Base error that is raised when the Sugar API responds with a failure status
    code.
    """

    def __init__(self, status_code: int, body: Any, *args: Any):
        try:
            error_message = str(body["error_message"])
        except TypeError:
            error_message = str(body)
        super(SugarError, self).__init__(error_message, *args)
        self.status_code = status_code


class InvalidSugarResponseError(ZuckerException):
    """Exceptions of this type are raised when the Sugar server sends an invalid
    response.
    """


class UnfetchedMetadataError(ZuckerException):
    """Raised when client metadata has not been fetched yet."""


class UnsavedRecordError(ZuckerException):
    """Raised when mutations are requested on a record that doesn't have a server-side
    ID yet.
    """



from typing import Mapping, Optional

from zucker.utils import JsonMapping, JsonType

from .base import AsyncClient


class AioClient(AsyncClient):
    """Asynchronous client implementation using `aiohttp`_.

    .. _aiohttp: https://docs.aiohttp.org/en/latest/index.html
    """

    def __init__(
        self,
        base_url: str,
        username: str,
        password: str,
        *,
        client_platform: str = "zucker",
        verify_ssl: bool = True,
    ):
        import aiohttp

        super().__init__(
            base_url,
            username,
            password,
            client_platform=client_platform,
            verify_ssl=verify_ssl,
        )

        self._session: Optional[aiohttp.ClientSession] = None

    async def close(self) -> None:
        if self._session is not None:
            await self._session.close()
            self._session = None

    async def raw_request(
        self,
        method: str,
        endpoint: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        data: Optional[JsonMapping] = None,
        json: Optional[JsonMapping] = None,
    ) -> tuple[int, JsonType]:
        import aiohttp

        if self._session is None:
            self._session = aiohttp.ClientSession()
            # The following doesn't actually do anything at the moment, but that might
            # change in a future version of aiohttp.
            await self._session.__aenter__()

        async with self._session.request(
            method,
            f"{self.base_url}/rest/v11_5/{endpoint}",
            headers={
                "OAuth-Token": self._authentication[1],
                "Cache-Control": "no-cache",
            },
            verify_ssl=self._verify_ssl,
            params=params or None,
            data=data or None,
            json=json or None,
        ) as response:
            return response.status, (await response.json())



from __future__ import annotations

from typing import Union
from urllib import parse as urllib_parse
from uuid import UUID

from zucker import filtering
from zucker.codegen.inspection import field_for_metadata
from zucker.utils import JsonType

from .base import MutableNumericField, MutableScalarField, ScalarField

__all__ = [
    "BooleanField",
    "FloatField",
    "IdField",
    "IntegerField",
    "LegacyEmailField",
    "StringField",
    "URLField",
]


# See this page for a reference of field types:
# https://support.sugarcrm.com/Documentation/Sugar_Versions/11.2/Pro/Administration_Guide/Developer_Tools/Studio/Fields/


@field_for_metadata.register(metadata_attributes=dict(name="id"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(type="id"), require_db=True)
class IdField(ScalarField[UUID, str]):
    """Immutable field for accessing record IDs.

    An instance of this field is automatically created for every module. Manually
    creating fields of this type should only be necessary for referencing links. In
    Python, IDs are represented by :class:`UUID` objects.
    """

    @classmethod
    def load_value(cls, raw_value: JsonType) -> UUID:
        if not isinstance(raw_value, str):
            raise TypeError(f"IDs must be strings, got {type(raw_value)!r}")
        return UUID(raw_value)

    @classmethod
    def serialize(cls, value: Union[UUID, str]) -> str:
        if isinstance(value, str):
            # "Load" the value to check if it is a valid ID.
            value = cls.load_value(value)
        return str(value)


@field_for_metadata.register(metadata_attributes=dict(type="url"), require_db=True)
class URLField(MutableScalarField[urllib_parse.ParseResult, str]):
    """Mutable string field for URLs.

    This should be used for database columns of type ``url``. URLs will be decoded and
    returned as an instance of :class:`urllib.parse.ParseResult`, which contains the
    parsed components of the URL. To get a string, call ``str()`` with the result
    object.
    """

    @staticmethod
    def load_value(raw_value: JsonType) -> urllib_parse.ParseResult:
        if not isinstance(raw_value, str):
            raise TypeError(
                f"URL field must be populated with a string - got "
                f"{type(raw_value)!r}"
            )
        return urllib_parse.urlparse(raw_value)

    @staticmethod
    def serialize(value: Union[urllib_parse.ParseResult, str]) -> str:
        parsed_value = (
            value
            if isinstance(value, urllib_parse.ParseResult)
            else URLField.load_value(value)
        )
        return parsed_value.geturl()


# Note: this field needs to be registered before StringField because of the matching
# order here:
@field_for_metadata.register(metadata_attributes=dict(name="email1"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(name="email2"), require_db=True)
class LegacyEmailField(MutableScalarField[str, str]):
    """Field for legacy email addresses (``email1`` and ``email2``).

    This is a string field that also validates [#validation]_ any input to be
    email-like. It should only be used on string columns with names ``email1`` and
    ``email2``.

    .. [#validation] Validation is currently pretty basic, because validating emails is hard. The
      Sugar backend will have the final call on what is treated as valid.
    """

    @staticmethod
    def load_value(raw_value: JsonType) -> str:
        if not isinstance(raw_value, str):
            raise TypeError(
                f"string field must be populated with a string - got "
                f"{type(raw_value)!r}"
            )

        # Email parsing is hard, and we don't really know how Sugar does it. To make it
        # easy, everything with an at is considered somewhat valid:
        if "@" not in raw_value:
            raise ValueError(f"invalid email: {raw_value}")

        return raw_value

    @staticmethod
    def serialize(value: str) -> str:
        return LegacyEmailField.load_value(value)


@field_for_metadata.register(metadata_attributes=dict(type="varchar"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(type="text"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(type="encrypt"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(type="longtext"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(type="textarea"), require_db=True)
class StringField(MutableScalarField[str, str]):
    """Mutable field that handles the various string columns.

    This will handle any backend field that has the database type ``varchar``, ``text``,
    ``text``, ``encrypt``, ``longtext`` or ``textarea``.
    """

    @staticmethod
    def load_value(raw_value: JsonType) -> str:
        if not isinstance(raw_value, str):
            raise TypeError(
                f"string field must be populated with a string - got "
                f"{type(raw_value)!r}"
            )
        return raw_value

    @staticmethod
    def serialize(value: str) -> str:
        return value

    def starts_with(self, prefix: str) -> filtering.StringFilter:
        """Filter for values that start with a given string."""
        return filtering.StringStartsFilter(self.name, prefix)

    def ends_with(self, suffix: str) -> filtering.StringFilter:
        """Filter for values that end with a given string."""
        return filtering.StringEndsFilter(self.name, suffix)

    def contains(self, infix: str) -> filtering.StringFilter:
        """Filter for values that contain a given string."""
        return filtering.StringContainsFilter(self.name, infix)

    def not_empty(self) -> filtering.NotEmptyFilter:
        """Filter for non-empty values."""
        return filtering.NotEmptyFilter(self.name)


@field_for_metadata.register(metadata_attributes=dict(type="bool"), require_db=True)
class BooleanField(MutableScalarField[bool, bool]):
    """Mutable field for boolean columns."""

    @staticmethod
    def load_value(raw_value: JsonType) -> bool:
        if not isinstance(raw_value, bool):
            raise TypeError(
                f"boolean field must be populated with a boolean - got "
                f"{type(raw_value)!r}"
            )
        return raw_value

    @staticmethod
    def serialize(value: bool) -> bool:
        return value

    def true(self) -> filtering.ValuesFilter[bool]:
        """Filter for true values."""
        return self.values(True)

    def false(self) -> filtering.ValuesFilter[bool]:
        """Filter for false values."""
        return self.values(False)


@field_for_metadata.register(metadata_attributes=dict(type="float"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(type="decimal"), require_db=True)
class FloatField(MutableNumericField[float]):
    """Mutable field for floating-point number columns.

    This is appropriate for backend fields that have the type ``float`` or ``decimal``.
    """

    @staticmethod
    def load_value(raw_value: JsonType) -> float:
        if not isinstance(raw_value, float):
            raise TypeError(
                f"float field must be populated with a float - got "
                f"{type(raw_value)!r}"
            )
        return raw_value

    @staticmethod
    def serialize(value: float) -> float:
        return value


@field_for_metadata.register(metadata_attributes=dict(type="int"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(type="integer"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(type="long"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(type="smallint"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(type="tinyint"), require_db=True)
@field_for_metadata.register(metadata_attributes=dict(type="ulong"), require_db=True)
class IntegerField(MutableNumericField[int]):
    """Mutable field for integer columns.

    Use this for backend fields of type ``int``, ``integer``, ``long``, ``smallint``,
    ``tinyint`` or ``ulong``.
    """

    @staticmethod
    def load_value(raw_value: JsonType) -> int:
        if not isinstance(raw_value, int):
            raise TypeError(
                f"integer field must be populated with an int - got "
                f"{type(raw_value)!r}"
            )
        return raw_value

    @staticmethod
    def serialize(value: int) -> int:
        return value



from __future__ import annotations

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

from zucker.client import SyncClient
from zucker.exceptions import InvalidSugarResponseError
from zucker.utils import JsonMapping, JsonPrimitive, JsonType

if TYPE_CHECKING:
    from zucker.model.fields.base import Field

    FieldInitializerReturnType = Optional[
        tuple[Type[Field[Any, Any]], Mapping[str, Any]]
    ]
    JsonPrimitiveOrCheckFn = Union[
        JsonPrimitive, Type[Any], Callable[[JsonPrimitive], bool]
    ]

_F = TypeVar("_F", bound="Field[Any, Any]", covariant=True)

__all__ = [
    "InspectedModule",
    "InspectedField",
    "field_for_metadata",
    "get_metadata",
    "inspect_modules_with_fields",
]


@dataclass
class InspectedModule:
    name: str
    fields: Sequence[InspectedField]
    class_arguments: Mapping[str, Any]
    raw_metadata: JsonMapping


@dataclass
class InspectedField:
    name: str
    field_type: Type[Field[Any, Any]]
    arguments: Mapping[str, Any]
    raw_metadata: JsonMapping


@dataclass
class FieldInspectionContext:
    field_metadata: JsonMapping
    modules: Mapping[str, InspectedModule]


class FieldMetadataRegistry:
    def __init__(self) -> None:
        self.field_initializers: list[
            Callable[[FieldInspectionContext], FieldInitializerReturnType]
        ] = []
        self.field_types: set[Type[Field[Any, Any]]] = set()

    def __call__(self, context: FieldInspectionContext) -> FieldInitializerReturnType:
        """Find the first registered field that accepts a specified context."""
        for initialize in self.field_initializers:
            result = initialize(context)
            if result is not None:
                return result
        return None

    def register(
        self,
        metadata_attributes: Optional[Mapping[str, JsonPrimitiveOrCheckFn]] = None,
        optional_metadata_attributes: Optional[
            Mapping[str, JsonPrimitiveOrCheckFn]
        ] = None,
        require_db: bool = False,
        output_arguments: Union[
            None,
            Mapping[str, Any],
            Callable[[FieldInspectionContext], Mapping[str, Any]],
        ] = None,
    ) -> Callable[[Type[_F]], Type[_F]]:
        """Register a server-side metadata construct that this field supports.

        The attribute options passed here are the ones you get when querying server
        metadata and looking in the 'fields' value in a module. The
        `Sugar Documentation`_ has a more detailed explanation of what information this
        object contains. These are passed in two parameters:

        - ``metadata_attributes`` contains attributes that the field must have in order
          to match. This is a dictionary where each value is either the direct attribute
          value or a callback that returns a boolean.
        - ``optional_metadata_attributes`` works the same way, but will also accept a
          field when the attribute isn't present at all (but will still reject when the
          value does not match).

        If the ``require_db`` parameter is true, the field will only match if the ``source``
        attribute is not ``non-db``.

        When inspection is run, any field that has attributes matching those given
        here will be exported as the field type that is decorated with this method.
        These attribute may either be given directly or
        The ``output_arguments`` option can be used to suggest field arguments for the
        newly created field.

        .. _Sugar Documentation: https://support.sugarcrm.com/Documentation/Sugar_Developer/Sugar_Developer_Guide_10.0/Data_Framework/Vardefs/#Fields_Array
        """

        def decorate(field_type: Type[_F]) -> Type[_F]:
            def initialize(
                context: FieldInspectionContext,
            ) -> FieldInitializerReturnType:
                def check_metadata_attribute(
                    key: str, expected_value: JsonPrimitiveOrCheckFn
                ) -> Optional[bool]:
                    # Go through the entire metadata tree and find the exact subtree we
                    # are looking for. This will make sure that when we are given a key
                    # of 'full_text_search.enabled' we are actually looking at the
                    # full_text_search subtree.
                    path_names = key.split(".")
                    current_item: Union[JsonMapping, JsonType] = context.field_metadata
                    for path_name in path_names:
                        if not isinstance(current_item, Mapping):
                            return None
                        if path_name not in current_item:
                            return None
                        current_item = current_item[path_name]
                    value = cast(JsonType, current_item)

                    # current_item should now be the expected value. The latter may
                    # either be given directly, as a type or as a callable.
                    if isinstance(expected_value, type):
                        if not isinstance(current_item, expected_value):
                            return False
                    elif callable(expected_value):
                        try:
                            if not expected_value(current_item):  # type: ignore
                                return False
                        except:
                            return False
                    else:
                        if current_item != expected_value:
                            return False
                    return True

                for item in (metadata_attributes or {}).items():
                    if check_metadata_attribute(*item) is not True:
                        return None
                for item in (optional_metadata_attributes or {}).items():
                    # Optional attributes must either match or not be present.
                    if check_metadata_attribute(*item) is False:
                        return None
                if require_db and (
                    check_metadata_attribute(
                        "source", lambda source: source != "non-db"
                    )
                    is False
                ):
                    return None

                # If we got until here, the field matches.
                suggested_arguments: Mapping[str, Any]
                if output_arguments is None:
                    suggested_arguments = {}
                elif callable(output_arguments):
                    try:
                        suggested_arguments = output_arguments(context)
                    except:
                        return None
                else:
                    suggested_arguments = output_arguments

                extra_arguments = {}
                if "name" in context.field_metadata:
                    extra_arguments["api_name"] = context.field_metadata["name"]

                return field_type, {**extra_arguments, **suggested_arguments}

            self.field_initializers.append(initialize)
            self.field_types.add(field_type)
            return field_type

        return decorate


field_for_metadata = FieldMetadataRegistry()


def inspect_modules_with_fields(
    modules_metadata: JsonMapping,
) -> Sequence[InspectedModule]:
    # TODO Convert these to actual type guards:

    def assert_sugar_mapping(mapping: Any) -> None:
        if not isinstance(mapping, Mapping):
            raise InvalidSugarResponseError(
                f"expecting mapping type, got {type(mapping)!r}"
            )

    def assert_sugar_contains(mapping: Any, name: str) -> None:
        assert_sugar_mapping(mapping)
        if name not in mapping:
            raise InvalidSugarResponseError(f"expecting mapping that contains {name!r}")

    # First pass: create an inspection result for each module. This allows all modules
    # to be referenced, even if their fields haven't been populated yet.
    modules: dict[str, InspectedModule] = {}
    for module_name, module_metadata in modules_metadata.items():
        if module_name in modules:
            raise InvalidSugarResponseError(
                f"got duplicate module name: {module_name!r}"
            )
        modules[module_name] = InspectedModule(
            name=module_name,
            fields=[],
            class_arguments=dict(
                api_name=module_name,
            ),
            raw_metadata=module_metadata,  # type: ignore
        )

    # Second pass: go through each module again and inspect their fields.
    for module_name, module_metadata in modules_metadata.items():
        assert_sugar_contains(module_metadata, "fields")
        assert_sugar_mapping(module_metadata["fields"])  # type: ignore

        fields = list[InspectedField]()

        for field_name, field_metadata in module_metadata["fields"].items():  # type: ignore
            if "name" in field_metadata and field_name != field_metadata["name"]:
                raise InvalidSugarResponseError(
                    f"non-matching field names {field_name!r} (from field set key) and "
                    f"{field_metadata['name']!r} (from actual field metadata)"
                )

            inspect_context = FieldInspectionContext(
                field_metadata=field_metadata,
                modules=modules,
            )
            inspect_result = field_for_metadata(inspect_context)
            if inspect_result is None:
                continue

            inspected_field = InspectedField(
                name=field_name,
                field_type=inspect_result[0],
                arguments=inspect_result[1],
                raw_metadata=field_metadata,
            )
            fields.append(inspected_field)

        modules[module_name].fields = fields

    return list(modules.values())
